"""
Spline Guide
César Vonc
CINEMA 4D R16
v 0.8 Novembre 2014
v 0.9 Mai 2015
v 1.0 Mai 2015
v 1.1 Juin 2015
"""

import sys
import os
import c4d
import math
from c4d import plugins, bitmaps, Vector, utils
from c4d.utils import RotAxisToMatrix, noise, FCut, SplineHelp, MixVec

MODULE_ID = 1033862

VONC_SPLINEGUIDE_FORCE = 1000
VONC_SPLINEGUIDE_SPLINE = 1001
VONC_SPLINEGUIDE_CHEVEUX = 1002
VONC_SPLINEGUIDE_INTENS = 1003
VONC_SPLINEGUIDE_RAYON = 1004
VONC_SPLINEGUIDE_SEG = 1005
VONC_SPLINEGUIDE_INFO = 1006
VONC_SPLINEGUIDE_SEGINIT = 1007
VONC_SPLINEGUIDE_SEGRELA = 1008
VONC_SPLINEGUIDE_SEGCACHE = 1009
VONC_SPLINEGUIDE_SEGDON = 1010
VONC_SPLINEGUIDE_DISPERSION = 1011
VONC_SPLINEGUIDE_DISPSPLINE = 1012
VONC_SPLINEGUIDE_TORSION = 1013
VONC_SPLINEGUIDE_TORSPLINE = 1014
VONC_SPLINEGUIDE_BRUIT = 1015
VONC_SPLINEGUIDE_BRUITECH = 1016
VONC_SPLINEGUIDE_BRUITSPLINE = 1017
VONC_SPLINEGUIDE_ADAP = 1018
VONC_SPLINEGUIDE_SEGMENT = 1019
VONC_SPLINEGUIDE_INVERSER = 1020

VERSION = "1.1"

class SplineGuide(plugins.ObjectData):
	"""Spline Guide"""
	
	def __init__(soi) :
		soi._force = 10
		soi._spline = None
		soi._cheveux = None
		soi._intens = 1.0
		soi._rayon = 100.0
		soi._seg = 0
		soi._guideDebut = []
		soi._guideLongSeg = []
		soi._cache = c4d.BaseContainer()
		soi._disp = 0
		soi._torsion = 0
		soi._bruit = 0
		soi._bruitech = 10.0
		soi._adapt = True
		soi._segment = 0
		soi._inverser = False
		
		soi._courbe = c4d.SplineData()
		soi._courbe.MakeLinearSplineBezier(2)
		soi._courbe.InitDefaultFlag(0)
		soi._courbe.SetKnot(index = 0, vPos = Vector(0.0), lFlagsSettings = 0, vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		soi._courbe.SetKnot(index = 1, vPos = Vector(1.0, 1.0, 0.0), lFlagsSettings = 0, vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		
		soi._torspline = c4d.SplineData()
		soi._torspline.MakeLinearSplineBezier(2)
		soi._torspline.InitDefaultFlag(0)
		soi._torspline.SetKnot(index = 0, vPos = Vector(0.0, 1.0, 0.0), lFlagsSettings = 0, vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		soi._torspline.SetKnot(index = 1, vPos = Vector(1.0, 1.0, 0.0), lFlagsSettings = 0, vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		
		soi._bruitspline = c4d.SplineData()
		soi._bruitspline.MakeLinearSplineBezier(2)
		soi._bruitspline.InitDefaultFlag(0)
		soi._bruitspline.SetKnot(index = 0, vPos = Vector(0.0, 1.0, 0.0), lFlagsSettings = 0, vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		soi._bruitspline.SetKnot(index = 1, vPos = Vector(1.0, 1.0, 0.0), lFlagsSettings = 0, vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		
	def GetHandleCount(op) :
		return 1
	
	def GetHandle(soi, op, i, info) :
		donnees = op.GetDataInstance()
		if donnees is None : return
		
		if i == 0 :
			info.position = Vector(donnees.GetFloat(VONC_SPLINEGUIDE_RAYON), 0.0, 0.0)
			info.direction = Vector(1.0, 0.0, 0.0)
			info.type = c4d.HANDLECONSTRAINTTYPE_LINEAR
	
	def DetectHandle(soi, op, bd, x, y, qualifier) :
		if qualifier & c4d.QUALIFIER_CTRL : return c4d.NOTOK
		
		mg = op.GetMg()
		ret = c4d.NOTOK
		
		i = 0
		info = c4d.HandleInfo()
		soi.GetHandle(op, i, info)
		if bd.PointInRange(info.position * mg, x, y) :
			ret = i
		
		return ret
	
	def MoveHandle(soi, op, undo, mouse_pos, hit_id, qualifier, bd) :
		dst = op.GetDataInstance()
		
		info = c4d.HandleInfo()
		val = mouse_pos.x
		soi.GetHandle(op, hit_id, info)
		
		if bd is not None :
			mg = op.GetUpMg() * undo.GetMl()
			pos = bd.ProjectPointOnLine(info.position * mg, info.direction ^ mg, mouse_pos.x, mouse_pos.y)
			val = (pos[0] * ~mg) * info.direction
		
		if hit_id == 0 :
			dst.SetFloat(VONC_SPLINEGUIDE_RAYON, FCut(val, 0.0, sys.maxint))
		
		return True
	
	def Draw(soi, op, drawpass, bd, bh) :
		dst = op.GetDataInstance()
		if drawpass == c4d.DRAWPASS_OBJECT :
			m = bh.GetMg()
			rayon = dst.GetFloat(VONC_SPLINEGUIDE_RAYON)
			m.v1 *= rayon
			m.v2 *= rayon
			m.v3 *= rayon
			bd.SetMatrix_Matrix(None, c4d.Matrix())
			bd.SetPen(bd.GetObjectColor(bh, op))
			bd.DrawCircle(m)
			h = m.v2
			m.v2 = m.v3
			m.v3 = h
			bd.DrawCircle(m)
			h = m.v1
			m.v1 = m.v3
			m.v3 = h
			bd.DrawCircle(m)
			return c4d.DRAWRESULT_OK
		
		elif drawpass == c4d.DRAWPASS_HANDLES :
			i = 0
			bd.SetMatrix_Matrix(None, bh.GetMg())
			
			surbrille = op.GetHighlightHandle(bd)
			
			info = c4d.HandleInfo()
			soi.GetHandle(op, i, info)
			if i == surbrille :
				bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_SELECTION_PREVIEW))
			else :
				bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT))
				
			bd.DrawHandle(info.position, c4d.DRAWHANDLE_BIG, 0)
			
			info = c4d.HandleInfo()
			soi.GetHandle(op, 1, info)
			bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT))
			bd.DrawLine(info.position, Vector(0), 0)
			
			return c4d.DRAWRESULT_OK
		
		return c4d.DRAWRESULT_SKIP
	
	def Init(soi, node) :
		donnees = node.GetDataInstance()
		donnees.SetFloat(VONC_SPLINEGUIDE_FORCE, soi._force)
		donnees.SetFloat(VONC_SPLINEGUIDE_INTENS, soi._intens)
		donnees.SetFloat(VONC_SPLINEGUIDE_RAYON, soi._rayon)
		donnees.SetFloat(VONC_SPLINEGUIDE_SEG, soi._seg)
		donnees.SetContainer(VONC_SPLINEGUIDE_SEGCACHE, soi._cache)
		donnees.SetFloat(VONC_SPLINEGUIDE_DISPERSION, soi._disp)
		donnees.SetData(VONC_SPLINEGUIDE_DISPSPLINE, soi._courbe)
		donnees.SetFloat(VONC_SPLINEGUIDE_TORSION, soi._torsion)
		donnees.SetData(VONC_SPLINEGUIDE_TORSPLINE, soi._torspline)
		donnees.SetFloat(VONC_SPLINEGUIDE_BRUIT, soi._bruit)
		donnees.SetFloat(VONC_SPLINEGUIDE_BRUITECH, soi._bruitech)
		donnees.SetData(VONC_SPLINEGUIDE_BRUITSPLINE, soi._bruitspline)
		donnees.SetInt32(VONC_SPLINEGUIDE_SEGMENT, soi._segment)
		donnees.SetBool(VONC_SPLINEGUIDE_INVERSER, soi._inverser)
		donnees.SetString(VONC_SPLINEGUIDE_INFO, "v " + VERSION + " - http://code.vonc.fr")
		
		return True
	
	def RecupCache(soi) :
		if not soi._cache or len(soi._cache) <= 1 :
			soi._cache = c4d.BaseContainer()
			return
		
		tai = len(soi._cache)
		tai2 = int(tai / 2)
		
		if soi._guideDebut and len(soi._guideDebut) == tai2 and len(soi._guideLongSeg) == tai2 : return
		
		soi._guideDebut = []
		soi._guideLongSeg = []
		
		for i in range(tai2) :
			soi._guideDebut.append(soi._cache[i])
		
		for i in range(tai2) :
			soi._guideLongSeg.append(soi._cache[i+tai2])
		
	
	def EnregistreCache(soi, donnees) :
		soi._cache = c4d.BaseContainer()
		tai = len(soi._guideDebut)
		if len(soi._guideLongSeg) != tai : return
		
		tai *= 2
		tab = []
		tab.extend(soi._guideDebut)
		tab.extend(soi._guideLongSeg)
		
		for i, v in enumerate(tab) :
			soi._cache[i] = v
		
		soi._seg = 0
		donnees.SetFloat(VONC_SPLINEGUIDE_SEG, soi._seg)
		donnees.SetContainer(VONC_SPLINEGUIDE_SEGCACHE, soi._cache)
		donnees.SetString(VONC_SPLINEGUIDE_SEGDON, str(len(soi._cache)))
	
	def PurgeCache(soi, donnees) :
		soi._cache = c4d.BaseContainer()
		donnees.SetContainer(VONC_SPLINEGUIDE_SEGCACHE, soi._cache)
		donnees.SetString(VONC_SPLINEGUIDE_SEGDON, str(len(soi._cache)))
	
	def RecupSel(soi, selCheveux, guideActif, ptsActif, guiSeg) :
		if not selCheveux : return
		bs = selCheveux.GetSelected()
		stype = selCheveux.GetSelectionType()
		nbGuides = len(guideActif)
		nbPts = guiSeg * nbGuides
		
		if stype == c4d.HAIR_MODE_GUIDES :
			tab = bs.GetAll(nbGuides)
			for i, sel in enumerate(tab) :
				if not sel : guideActif[i] = False
			return False
		
		elif stype == c4d.HAIR_MODE_TIPS or stype == c4d.HAIR_MODE_POINTS :
			tab = bs.GetAll(nbPts)
			for i, sel in enumerate(tab) :
				if sel : ptsActif[i] = True
			return True
			
		elif stype == c4d.HAIR_MODE_ROOTS : # Racines
			tab = bs.GetAll(nbPts)
			for i, g in enumerate(guideActif) :
				if not g : continue
				if not bs.IsSelected(i * guiSeg) :
					guideActif[i] = False
			return False
		
		return False
	
	def RecupCheveux(soi, obj) :
		if not obj : return None, None
		if obj.CheckType(c4d.Thairselection) :
			return obj.GetObject(), obj
		return obj, None
		
	def ModifyParticles(soi, op, pp, ss, pcnt, diff):
		donnees = op.GetDataInstance()
		if not donnees : return
		
		soi._force = donnees.GetFloat(VONC_SPLINEGUIDE_FORCE)
		soi._intens = donnees.GetFloat(VONC_SPLINEGUIDE_INTENS)
		soi._spline = donnees.GetLink(VONC_SPLINEGUIDE_SPLINE)
		soi._cheveux, selCheveux = soi.RecupCheveux(donnees.GetLink(VONC_SPLINEGUIDE_CHEVEUX))
		soi._rayon = donnees.GetFloat(VONC_SPLINEGUIDE_RAYON)
		soi._seg = donnees.GetFloat(VONC_SPLINEGUIDE_SEG)
		soi._cache = donnees.GetContainer(VONC_SPLINEGUIDE_SEGCACHE)
		soi._disp = donnees.GetFloat(VONC_SPLINEGUIDE_DISPERSION) * .5
		soi._courbe = donnees.GetData(VONC_SPLINEGUIDE_DISPSPLINE)
		soi._torsion = donnees.GetFloat(VONC_SPLINEGUIDE_TORSION)
		soi._torspline = donnees.GetData(VONC_SPLINEGUIDE_TORSPLINE)
		soi._bruit = donnees.GetFloat(VONC_SPLINEGUIDE_BRUIT)
		soi._bruitech = donnees.GetFloat(VONC_SPLINEGUIDE_BRUITECH)
		soi._bruitspline = donnees.GetData(VONC_SPLINEGUIDE_BRUITSPLINE)
		soi._segment = donnees.GetInt32(VONC_SPLINEGUIDE_SEGMENT)
		soi._inverser = donnees.GetInt32(VONC_SPLINEGUIDE_INVERSER)
		donnees.SetString(VONC_SPLINEGUIDE_SEGDON, str(len(soi._cache)))
		donnees.SetString(VONC_SPLINEGUIDE_INFO, "v " + VERSION + " - http://code.vonc.fr")
		
		if not soi._spline : return
		if not soi._cheveux : return
		
		
		cerAide = SplineHelp()
		if not cerAide.InitSpline(soi._spline, Vector(), None, True, True, False, True) : return
		soi._segment = soi._segment % cerAide.GetSegmentCount()
		cerLong = cerAide.GetSegmentLength(soi._segment)
		cerPos = soi._spline.GetMg().off
		guides = soi._cheveux.GetDynamicGuides()
		guiSeg = guides.GetGuidePointCount()
		
		soi.RecupCache()
		
		intens = 1.0 - (soi._intens ** 2)
		fac = diff * soi._force * 100.
		rayon = soi._rayon ** 2
		lpp = len(pp)
		pos = op.GetMg().off
		i = 0
		j = 0
		
		seginit = not not soi._cache
		
		decalage = soi._seg
		guideDmin = 0
		guideLong = 0
		guideActif = []
		
		if not seginit :
			soi._guideDebut = []
			soi._guideLongSeg = []
			decalage = 0
		
		if not soi._adapt :
			soi._guideDebut = []
			soi._guideLongSeg = []
			decalage = soi._seg
			seginit = False
		
		if soi._inverser :
			guidePos = cerAide.GetPosition(1,soi._segment,True,True)
		else :
			guidePos = cerAide.GetPosition(0,soi._segment,True,True)
		
		g = 0
		
		__long = (pp[0].off - pp[(0+1)%lpp].off).GetLength()
		
		for i, s in enumerate(pp) :
			if not (s.bits & c4d.PARTICLEFLAGS_VISIBLE) : continue
			
			j = i % guiSeg
			guideLong = (s.off - guidePos).GetLengthSquared()
			
			if j == 0 :
				g += 1
				guideDmin = guideLong
				if (s.off - pos).GetLengthSquared() > rayon : guideActif.append(False)
				else : guideActif.append(True)
				if not seginit :
					soi._guideDebut.append(0)
					if soi._adapt :
						soi._guideLongSeg.append((s.off - pp[(i+1)%lpp].off).GetLength())
					else :
						soi._guideLongSeg.append(__long)
			else :
				if guideLong < guideDmin :
					guideDmin = guideLong
					if not seginit :
						if soi._adapt :
							soi._guideDebut[-1] = j
						else :
							soi._guideDebut[-1] = 0
		
		if len(soi._guideDebut) != g :
			soi.PurgeCache(donnees)
			cerAide.FreeSpline()
			return
		
		# Sauvegarde Fixe zéro
		if not soi._adapt :
			soi._adapt = True
			soi.EnregistreCache(donnees)
		#
		
		ptsActif = [False] * (len(guideActif) * guiSeg)
		ptsActifBool = soi.RecupSel(selCheveux, guideActif, ptsActif, guiSeg)
		
		i = 0
		g = -1
		pct = 0
		
		
		for s, p in zip(pp, ss):
			if not (s.bits & c4d.PARTICLEFLAGS_VISIBLE) :
				i += 1
				continue
			
			j = i % guiSeg
			if (j == 0) : g += 1
			guiDeb = soi._guideDebut[g]
			guiAct = guideActif[g]
			ptsAct = True
			if ptsActifBool : ptsAct = ptsActif[i]
			guiLongSeg = soi._guideLongSeg[g]
			
			if not guiAct or not ptsAct : pct = 2
			else : pct = ((j-guiDeb) * guiLongSeg) / cerLong + decalage
			
			if pct <= 1 and pct >= 0 :
				if soi._inverser : pct = 1.0 - pct
				pos = cerAide.GetPosition(pct,soi._segment,True,True)
				
				if soi._disp :
					norrot = (float(i) / lpp) * 6.28318530718
					
					dispval = soi._disp * soi._courbe.GetPoint(pct).y
					
					if soi._torsion :
						norrot += pct * soi._torspline.GetPoint(pct).y * soi._torsion
					
					norx = cerAide.GetTangent(pct, soi._segment, True, True).GetNormalized()
					nory = cerAide.GetCrossNormal(pct, soi._segment, True, True).GetNormalized()
					sn, cs = utils.SinCos(norrot)
					nor = norx * sn + nory * cs
					
					nor *= soi._disp * soi._courbe.GetPoint(pct).y
					
					pos = pos + nor
					
					if soi._bruit :
						_bruit_i = soi._bruitspline.GetPoint(pct).y * soi._bruit * 10.
						_bruit_x = noise.SNoise(Vector(pct*6., norrot, 0) * soi._bruitech) * dispval * _bruit_i
						_bruit_y = noise.SNoise(Vector(pct*6. + 100, norrot, 0) * soi._bruitech) * dispval * _bruit_i
						pos += norx * _bruit_x
						pos += nory * _bruit_y
					
				
				vec = pos - s.off
				
				if soi._intens != 1.0 :
					p.v = MixVec(vec * fac, s.v3, intens)
				else :
					p.v = vec * fac
				
				p.count += 1
			
			i += 1
		
		cerAide.FreeSpline()
		return
	
	def Message(soi, op, type, donnees) :
		if type == c4d.MSG_DESCRIPTION_COMMAND :
			id = donnees["id"][0].id
			if id == VONC_SPLINEGUIDE_SEGINIT : soi.EnregistreCache(op.GetDataInstance())
			elif id == VONC_SPLINEGUIDE_SEGRELA : soi.PurgeCache(op.GetDataInstance())
			elif id == VONC_SPLINEGUIDE_ADAP : soi._adapt = False
		return True
	
	def GetDEnabling(soi, node, id, t_data, flags, itemdesc):
		data = node.GetDataInstance()
		if data is None: return
		
		if id[0].id == VONC_SPLINEGUIDE_SEG :
			if not data[VONC_SPLINEGUIDE_SEGCACHE] : return False
		
		elif id[0].id == VONC_SPLINEGUIDE_SEGINIT :
			if data[VONC_SPLINEGUIDE_SEGCACHE] : return False
		
		elif id[0].id == VONC_SPLINEGUIDE_ADAP :
			if data[VONC_SPLINEGUIDE_SEGCACHE] : return False
		
		return True


if __name__ == "__main__":
	path, fn = os.path.split(__file__)
	bmp = bitmaps.BaseBitmap()
	bmp.InitWith(os.path.join(path, "res", "icones", "Ovoncsplineguide.tif"))
	plugins.RegisterObjectPlugin(id=MODULE_ID, str="Spline Guide",
								g=SplineGuide,
								description="Ovoncsplineguide", icon=bmp,
								info=c4d.OBJECT_PARTICLEMODIFIER)
